import pygame
import sys
import math

import random

import ode

from BkGlobals import *

def setupScreen(width = 1024, height = 768, bWindowed = False):
	global SCREEN_X
	global SCREEN_Y
	global MAX_BLOCK_SIZE_X
	global MAX_BLOCK_SIZE_Y
	global BORDER_WIDTH_X
	global BORDER_WIDTH_Y
	global GAME_SURFACE_POS
	global GAME_SURFACE_SIZE
	global NEXT_BLOCK_SURFACE_POS
		
	if not width or not height:
		dispModes = pygame.display.list_modes()
		dispModes.sort(reverse = True)
	
		width = dispModes[0][0]
		height = dispModes[0][1]
		
	pygame.display.set_caption("PyBuckle")
	
	if bWindowed:
		pygame.display.set_mode((width, height))
	else:
		pygame.display.set_mode((width, height), pygame.FULLSCREEN)
	
	BORDER_WIDTH_X = BORDER_WIDTH_X * width / SCREEN_X
	BORDER_WIDTH_Y = BORDER_WIDTH_Y * height / SCREEN_Y
		
	MAX_BLOCK_SIZE_X = MAX_BLOCK_SIZE_X * width / SCREEN_X
	MAX_BLOCK_SIZE_Y = MAX_BLOCK_SIZE_Y * height / SCREEN_Y
					
	GAME_SURFACE_POS = (BORDER_WIDTH_X, BORDER_WIDTH_Y)
	GAME_SURFACE_SIZE = ((3 * width / 4) - BORDER_WIDTH_X, height - (2 * BORDER_WIDTH_Y))
	NEXT_BLOCK_SURFACE_POS = (width - MAX_BLOCK_SIZE_X - BORDER_WIDTH_X, BORDER_WIDTH_Y)
		
	SCREEN_X = width
	SCREEN_Y = height

			
def runGame():
	setupScreen()
	pygame.display.get_surface().fill(BG_COLOUR)
	
	world = ode.World()
	world.setGravity((0, -150 ,0))
	#world.setERP(0.2)
	#world.setContactSurfaceLayer(0.001)
	#world.setContactMaxCorrectingVel(100)
	world.setCFM(10E-10)	
	space = ode.Space()
	
	
	currentBlock = Block(world, space)
	
	gameArea = GameArea(world, space)
	#nextBlockArea = NextBlockArea()
	#nextBlockArea.update(currentBlock)
	
	frameRateClock = pygame.time.Clock()
	frameRateClock.tick()
		
	pygame.display.flip()	
		
	bRun = True
	iTick = 0
	iTime = 0
	iOldTime = 0
	while bRun:
		iOldTime = iTime
		iTime += frameRateClock.tick(GFX_TICK_RATE)
		updateRectList = []
	
		for event in pygame.event.get():
			if event.type == pygame.QUIT:
				bRun = False
			elif event.type == pygame.KEYDOWN:
				if event.dict["key"] == pygame.K_ESCAPE:
					bRun = False
				elif event.dict["key"] == pygame.K_p or event.dict["key"] == pygame.K_PAUSE:
					pass			
			elif event.type == pygame.MOUSEBUTTONUP:
				if event.dict["button"] == 1:
					pos = event.dict["pos"]
					
					if gameArea.setBlock(currentBlock, pos):
						currentBlock = Block(world, space)
						
		gameArea.updateMousePos(pygame.mouse.get_pos(), currentBlock)						
						
				
		gameAreaUpdateRect = gameArea.update((iTime - iOldTime) / 1000.0)
		if gameAreaUpdateRect:
			updateRectList.append(gameAreaUpdateRect)
				
		if updateRectList:
			pygame.display.update(updateRectList)
			
		iTick += 1
		

		
class NextBlockArea:
	def __init__(self):
		self.rows = [[Square((i, j)) for i in range(7)] for j in range(7)]
	
		self.nextBlockSurf = pygame.Surface((int(MAX_BLOCK_SIZE_X * 1.1), int(MAX_BLOCK_SIZE_Y * 1.1))).convert()
		self.nextBlockSurf.fill(NEXT_BLOCK_BG_COLOUR)
		
	def update(self, nextBlock):
		for row in self.rows:
			for square in row:
				square.setFull(False)
	
		for i in range(3 - (nextBlock.getWidth() / 2),  4 + (nextBlock.getWidth() / 2)):
			for j in range(3 - (nextBlock.getHeight() / 2),  4 + (nextBlock.getHeight() / 2)):
				self.rows[j][i].setFull(True)
	
		for row in self.rows:
			for square in row:
				if square.update():
					self.nextBlockSurf.blit(square.getSurface(), ((square.getX() + 1) * BLOCK_SIZE_X, (square.getY() + 1) * BLOCK_SIZE_Y))
			
		pygame.display.get_surface().blit(self.nextBlockSurf, NEXT_BLOCK_SURFACE_POS)
		
class TimerArea:
	def __init__(self):
		self.TimerSurf = pygame.Surface((BLOCK_SIZE_X * 9, (BLOCK_SIZE_Y * GAME_HEIGHT - BLOCK_SIZE_Y * 10)) / 2).convert()
						
class ScoreArea:				
	def __init__(self):
		self.TimerSurf = pygame.Surface((BLOCK_SIZE_X * 9, (BLOCK_SIZE_Y * GAME_HEIGHT - BLOCK_SIZE_Y * 10)) / 2).convert()
			
		
class GameArea:
	def __init__(self, world, space):
		# ODE
		self.space = space
		self.world = world
		self.contactgroup = ode.JointGroup()
		self.floor = ode.GeomPlane(space, (0, 1, 0), SCREEN_Y - (5 * GAME_SURFACE_SIZE[1] / 6))
		
		# Not ODE
		#self.bBlit = True
			
		self.blockList = []
		
		self.edgeRects = [(0, BORDER_WIDTH_X + (5 * GAME_SURFACE_SIZE[1] / 6), GAME_SURFACE_SIZE[0], 5 * GAME_SURFACE_SIZE[1] / 6), # Grass
						  (BORDER_WIDTH_X + GAME_SURFACE_SIZE[0], 0, SCREEN_X - BORDER_WIDTH_X + GAME_SURFACE_SIZE[0], SCREEN_Y), # Right panel
						  (0, 0, BORDER_WIDTH_X, SCREEN_Y), # Left border
						  (0, 0, SCREEN_X, BORDER_WIDTH_Y), # Top border
						  (0, SCREEN_Y - BORDER_WIDTH_Y, SCREEN_X, BORDER_WIDTH_Y)] # Bottom border
	
		self.baseSurface = pygame.Surface(GAME_SURFACE_SIZE).convert()
		self.baseSurface.fill((0, 0, 128))
		
		self.baseSurface.fill((0, 128, 0), (0, 5 * GAME_SURFACE_SIZE[1] / 6, GAME_SURFACE_SIZE[0], 5 * GAME_SURFACE_SIZE[1] / 6))
				
		self.gameSurface = self.baseSurface.copy()
				
	def snapToNearestValidPos(self, pos, block):
		#TODO - behaves oddly in bottom left, and can't handle diagonals (collisions with 2 rects won't give closest point)
		# Bot left behaviour probably snapping left (off screen)
	
		blockRect = block.getSurface().get_rect()
		blockRect.left -= blockRect.width / 2
		blockRect.top -= blockRect.height / 2
		
		blockRect.left += pos[0]
		blockRect.top += pos[1]
		
		
		rectNoCollideList = [] + self.edgeRects

		for newBlock in self.blockList:
			newBlockRect = newBlock.getSurface().get_rect()
			newBlockRect.left += newBlock.getPosition()[0]
			newBlockRect.top += newBlock.getPosition()[1]			
			
			rectNoCollideList.append(newBlockRect)
		
		testRect = blockRect
		if blockRect.collidelist(rectNoCollideList) == -1:
			# Find nearest
			bBreak = False
			testRect = blockRect
			
			for iIteration in range(1, max(GAME_SURFACE_SIZE)):			
				for i in range(-1 * iIteration, 2 * iIteration, iIteration):
					for j in range(-1 * iIteration, 2 * iIteration, iIteration):
						if abs(i) != abs(j):
							tmpTestRect = blockRect.move(i, j)
							if testRect.collidelist(rectNoCollideList) != -1:
								bBreak = True
								break
							testRect = tmpTestRect
							
					if bBreak:
						break
				if bBreak:
					break
					
			blockRect = testRect
										
		if blockRect.collidelist(rectNoCollideList) != -1:
			# bump away
			bBreak = False
			for iIteration in range(1, max(GAME_SURFACE_SIZE)):			
				for i in range(-1 * iIteration, 2 * iIteration, iIteration):
					for j in range(-1 * iIteration, 2 * iIteration, iIteration):
						if abs(i) != abs(j):
							testRect = blockRect.move(i, j)
							if testRect.collidelist(rectNoCollideList) == -1:
								bBreak = True
								break
					if bBreak:
						break
				if bBreak:
					break	
		
		if bBreak:
			return (testRect.left, testRect.top)
			
		return False

		
				
	def updateMousePos(self, pos, currentBlock):	
		self.mousePos = pos
		self.mouseBlock = currentBlock
						
						
	def drawMouse(self):
		if self.gameSurface.get_rect().collidepoint(self.mousePos):
			pygame.mouse.set_visible(False)
		else:
			return	
		blockSurf = self.mouseBlock.getSurface()
		blockRect = blockSurf.get_rect()
		
		newPos = self.snapToNearestValidPos(self.mousePos, self.mouseBlock)
				
		if newPos:
			self.gameSurface.blit(blockSurf, newPos)
			
		pygame.draw.circle(self.gameSurface, (128, 128, 0), self.mousePos, 5)
		#self.bBlit = True
		
	def setBlock(self, block, pos):
		if self.gameSurface.get_rect().collidepoint(pos):
			newPos = self.snapToNearestValidPos(pos, block)
		
			if newPos:
				block.setInitialPosition(newPos)
				self.blockList.append(block)
				
				#for block in self.blockList:
				#	self.gameSurface.blit(block.getSurface(), block.getPosition())
										
				#self.bBlit = True
				
				return True
			
		return False
		
	def update(self, dt):
	
		n = 10
		for i in range(n):
			self.space.collide((self.world, self.contactgroup), near_callback)
			self.world.step(dt / n)
			self.contactgroup.empty()
			
		blockKillList = []
			
		for block in self.blockList:
			if block.update():
				blockKillList.append(block)
				continue
		
			rect = block.getSurface().get_rect()
			bodyPos = block.getBody().getPosition()

			block.setPosition(((bodyPos[0] - rect.center[0]), (SCREEN_Y - (bodyPos[1] + rect.center[1]))))
			
			self.gameSurface.blit(block.getSurface(), block.getPosition())
								
		for block in blockKillList:
			self.blockList.remove(block)
			
		self.drawMouse()
								
		self.updateRect = pygame.display.get_surface().blit(self.gameSurface, (BORDER_WIDTH_X, BORDER_WIDTH_Y))
		#self.bBlit = False
			
		self.gameSurface = self.baseSurface.copy()
		
		return self.updateRect
			
		#self.gameSurface = self.baseSurface.copy()	
			
		#return False
						
			
class Block:
	def __init__(self, world, space):
		if not random.randint(0, 3):
			self.height = random.randint(1, 6) * (MAX_BLOCK_SIZE_Y / 6)
		elif not random.randint(0, 2):
			self.height = random.randint(1, 3) * (MAX_BLOCK_SIZE_Y / 6)
		elif not random.randint(0, 1):
			self.height = random.randint(1, 2) * (MAX_BLOCK_SIZE_Y / 6)
		else:
			self.height = MAX_BLOCK_SIZE_Y / 6
			
		if not random.randint(0, 3):
			self.width = random.randint(1, 6) * (MAX_BLOCK_SIZE_X / 6)
		elif not random.randint(0, 2):
			self.width = random.randint(1, 3) * (MAX_BLOCK_SIZE_X / 6)
		elif not random.randint(0, 1):
			self.width = random.randint(1, 2) * (MAX_BLOCK_SIZE_X / 6)
		else:	  
			self.width = MAX_BLOCK_SIZE_X / 6
			
				 		
		self.surf = pygame.Surface((self.width, self.height)).convert_alpha()
		
		self.surf.fill((0, 0, 0))
		self.surf.fill((255, 0, 0), (1, 1, self.width - 2, self.height - 2))
		
		self.pos = [0, 0]
		
		self.world = world
		self.space = space
		
		self.body = None
		self.iKillTimer = -1

	def kill(self):
		if self.iKillTimer == -1:
			self.iKillTimer = 0

	def getBody(self):
		return self.body
		
	def getSurface(self):
		if self.body:
			rotation = math.atan2(clamp(self.body.getRotation()[1], (-1, 1)), clamp(self.body.getRotation()[0], (-1, 1)))
			
			rotation *= 180 / math.pi
						
			rotation = -rotation
			
			if math.fabs(rotation) > 10:
				self.kill()
									
			surf = pygame.transform.rotate(self.surf, rotation)
		else:
			surf = self.surf
			
	
		return surf
		
	def getPosition(self):
		return self.pos
		
	def setInitialPosition(self, pos):
		rect = self.surf.get_rect()
	
		# Create body
		self.body = ode.Body(self.world)
		self.body.setPosition(((pos[0] + rect.center[0]), (SCREEN_Y - (pos[1] + rect.center[1])), 0))
		M = ode.Mass()
		M.setBox(BLOCK_DENSITY / 500.0, self.width, self.height, 500)
		self.body.setMass(M)

		# Create a box geom for collision detection
		self.geom = ode.GeomBox(self.space, lengths = (self.width, self.height, 500))
		self.geom.setBody(self.body)	
		
		self.pos = pos
		
	def setPosition(self, pos):
		self.pos = (pos[0], pos[1])
		
	def update(self):
		if self.iKillTimer >= BLOCK_KILL_TIME:
			self.body.disable()
			self.space.remove(self.geom)
			return True
		if self.iKillTimer != -1:
			self.iKillTimer += 1 
			self.surf.fill((0, 0, 0, 255 - 255 * self.iKillTimer / BLOCK_KILL_TIME))
			self.surf.fill((255, 0, 0, 255 - 255 * self.iKillTimer / BLOCK_KILL_TIME), (1, 1, self.width - 2, self.height - 2))
			
		return False
		
		
def clamp(value, limits):
	if value < limits[0]:
		value = limits[0]
	if value > limits[1]:
		value = limits[1]
		
	return value
		
# Collision callback
def near_callback(args, geom1, geom2):
	"""Callback function for the collide() method.

	This function checks if the given geoms do collide and
	creates contact joints if they do.
	"""

	# Check if the objects do collide
	contacts = ode.collide(geom1, geom2)

	# Create contact joints
	world,contactgroup = args
	for c in contacts:
		c.setBounce(0.05)
		c.setBounceVel(100)
		c.setMu(250000000)
		j = ode.ContactJoint(world, contactgroup, c)
		j.attach(geom1.getBody(), geom2.getBody())		
		
		
if __name__ == "__main__":
	pygame.init()
	
	import profile	
	profile.run("runGame()")
		
		